import sys, os, socket, traceback, select, signal
from datetime import datetime

from webob.response import iter_close
from pasteob import *
from pasteob.server.request import parse_request, CloseConnection#, SockRFile

from mext.threads import Thread, Event, Queue
from mext.log import Logger
from mext.context import State

log = Logger(__name__, min_level='INFO')

__all__ = ['WsgiServer']


class ServerExit(SystemExit):
    pass

is_win = (os.name == 'nt')

# @@ support chunked encoding

def new_ipv4_socket():
    return socket.socket()
    #return socket._realsocket(socket.AF_INET, socket.SOCK_STREAM)

class WsgiServer(object):
    request_queue_size = 20
    allow_reuse_address = False
    socket_timeout = None
    daemon_threads = False
    server_header = 'PasteOb/0.2'
    worker_count = 3

    @classmethod
    def serve(cls, app, host=None, port=8080):
        host = host or '127.0.0.1'
        server = cls(app, (host, port))
        log.info("Serving on %s:%s", host, port)
        try:
            server.do_serve()
        except KeyboardInterrupt:
            pass

    #
    # initial setup
    #

    def __init__(self, app, server_address):
        self.app = app
        self.state = State.get()
        self.server_address = server_address
        self._serving = False
        self.idle_conns = [] # a list of connections in keep-alive state
        self.create_socket()

    def create_socket(self):
        self.socket = new_ipv4_socket()
        if self.allow_reuse_address:
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind(self.server_address)
        self.socket.listen(self.request_queue_size)
        self.server_address = self.socket.getsockname()
        host, port = self.socket.getsockname()[:2]
        self.server_name = socket.getfqdn(host)
        self.server_port = port

    def create_wakeup_socket(self):
        sock = new_ipv4_socket()
        sock.bind(('127.0.0.1', 0))
        sock.listen(1)
        addr = sock.getsockname()
        wkw = new_ipv4_socket()
        wkw.connect(addr)
        wkr, _ = sock.accept()
        self._wakeup_w = wkw
        self._wakeup_r = wkr

    def start_workers(self):
        self.workers = []
        for i in xrange(self.worker_count):
            self._start_worker()

    def _start_worker(self):
        th = Thread.make(self.worker_loop)
        th.daemon = self.daemon_threads
        th.start()
        self.workers.append(th)





    #
    # eventloop
    #

    def wakeup(self):
        #log.debug("Request wakeup")
        self._wakeup_w.send('x')

    def do_serve(self, poll_interval=None):
        if poll_interval is None:
            poll_interval = 0.5 if is_win else 5
        self.create_wakeup_socket()
        self.req_queue = Queue()
        self.start_workers()
        self.setup_signals()
        self._serving = True

        try:
            while True:
                srv = self.socket
                wkr = self._wakeup_r
                r = [srv, wkr] + self.idle_conns
                try:
                    r, w, e = select.select(r, [], [], poll_interval)
                except select.error, exc:
                    if exc[0] == 4: # signal
                        continue
                    raise
                if not self._serving:
                    return
                for sock in r:
                    if sock is wkr:
                        #log.debug("Woke up")
                        wkr.recv(100) # consume all wakeup requests
                    elif sock is srv:
                        log.debug("New connection")
                        try:
                            conn, caddr = srv.accept()
                            conn.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
                            if self.socket_timeout:
                                conn.settimeout(self.socket_timeout)
                        except socket.error:
                            continue
                        self.process_request(conn)
                    else:
                        self.idle_conns.remove(sock)
                        #log.debug("New request")
                        self.process_request(sock)
        finally:
            log.info("Shutting down")
            self._serving = False
            self.restore_signals()
            log.debug("Closing sockets")
            self.close_sockets()
            log.debug("Requesting workers to stop")
            for i in xrange(len(self.workers)):
                self.req_queue.put(None) # quit request
            self.wait_workers()
            log.debug("Shutdown successful")


    def setup_signals(self):
        self._set_signals(self.on_term)

    def restore_signals(self):
        self._set_signals(signal.SIG_DFL)

    def close_sockets(self):
        socks = [self.socket, self._wakeup_r, self._wakeup_w]
        socks += self.idle_conns
        for sock in socks:
            sock.close()

    def _set_signals(self, handler):
        if is_win:
            signal.signal(signal.SIGBREAK, handler)
        signal.signal(signal.SIGTERM, handler)

    def on_term(self, sig, fr):
        log.info("Got TERM signal")
        self.shutdown()

    def wait_workers(self):
        assert Thread.get_current() not in self.workers
        for th in self.workers[:]:
            th.join()

    def shutdown(self):
        if self._serving:
            log.info("Requesting shutdown")
            self._serving = False
            self.wakeup()

    def process_request(self, conn):
        self.req_queue.put(conn)



    #
    # worker implementation
    #

    def worker_loop(self):
        try:
            while True:
                conn = self.req_queue.get()
                if conn is None: # quit
                    break
                self.handle_req(conn)
        finally:
            th = Thread.get_current()
            self.workers.remove(th)

    def handle_req(self, conn):
        try:
            #caddr = conn.getpeername()
            self._handle_req(conn)
        except socket.error, exc:
            pass
        except CloseConnection:
            #log.debug("Close connection to %s:%d", *caddr)
            conn.close()
            #SockRFile.delete_from(conn)
        except:
            log.error("Unexpected worker error", exc_info=True)
        else:
            self.idle_conns.append(conn)
            self.wakeup() # make sure conn participates in select()

    def _handle_req(self, conn):
        caddr = conn.getpeername()
        #log.debug("Activity from %s:%d", *caddr)
        env = resp = None
        http10 = True
        with self.state.child():
            try:
                env = parse_request(conn)
                env.update({
                    'pasteob.server': self,
                    'SERVER_NAME': self.server_address[0],
                    'SERVER_PORT': self.server_address[1],
                })
                http10 = (env['SERVER_PROTOCOL'] == 'HTTP/1.0')
                # we use env copy so that if apps change the env (for example method)
                resp = Request(env.copy()).get_response(self.app)
            except HTTPError, exc:
                disconnect = True
                resp = exc
            except CloseConnection:
                raise
            except socket.error:
                raise CloseConnection
            except:
                disconnect = True
                resp = HTTPServerError()
                log.error("Error handling request from %s", caddr[0], exc_info=True)
            else:
                disconnect = decide_disconnect(env, resp)
            resp = filter_response(env, resp)
            if not resp.server:
                resp.server = self.server_header
            if http10 and not disconnect:
                resp.headers['connection'] = 'keep-alive'
            elif not http10 and disconnect:
                resp.headers['connection'] = 'close'
            send_response(resp, conn)
        # dont-clear is a flag set by pasteob.debug.eval so the data is there
        # for the debugging
        #if req and 'dont-clear' not in req.environ:
        #    req.environ.clear() # remove cycles created by ReqAddons
        if disconnect:
            raise CloseConnection


################
################
################


def filter_response(env, resp):
    if isinstance(resp, HTTPError):
        if env is None:
            req = Request.blank('/')#, accept='text/html')
        else:
            req = Request(env) # this assumes resp does not modify env
        resp = req.get_response(resp)
        del req
    status_int = resp.status_int
    if ((env and env['REQUEST_METHOD'] == 'HEAD')
        or status_int < 200
        or status_int in (204, 304)
    ):
        iter_close(resp.app_iter)
        resp.app_iter = ()
    return resp


def send_response(resp, conn):
    if not resp.date:
        resp.date = datetime.now()
    if resp.content_length is None:
        resp.headers['connection'] = 'close'

    headers = ['HTTP/1.1 ' + resp.status]
    for k, v in resp.headerlist:
        headers.append('%s: %s' % (k, v))
    headers += ('', '')
    conn.sendall('\r\n'.join(headers))
    try:
        for chunk in resp.app_iter:
            conn.sendall(chunk)
    finally:
        iter_close(resp.app_iter)

def decide_disconnect(env, resp):
    req_conn = env.get('HTTP_CONNECTION', '').lower()
    resp_conn = resp.headers.get('connection', '').lower()
    if req_conn == 'close' or resp_conn == 'close':
        return True
    elif resp.status_int == 413: # request entity too large
        return True
    elif (resp.content_length is None
        and env['REQUEST_METHOD'] != 'HEAD'
        and resp.status_int >= 200
        and resp.status_int not in (204, 304)
    ):
        #@@ if http/1.1, convert response to chunked
        return True
    elif req_conn == 'keep-alive':
        return False
    else:
        return (env['SERVER_PROTOCOL'] == 'HTTP/1.0')






if __name__ == '__main__':
    log.min_level = 'debug'
    from pprint import pformat
    @wapp
    def dump_req(req):
        #print req
        if req.query_string == 'shutdown':
            req.environ['pasteob.server'].shutdown()
        return Response(str(req)+'\n'+pformat(req.environ), content_type='text/plain')
    WsgiServer.serve(dump_req)

